/*
 * Copyright (c) 2021 Nordic Semiconductor ASA
 *
 * SPDX-License-Identifier: LicenseRef-Nordic-5-Clause
 */

#include <zephyr/kernel.h>
#include <zephyr/sys/printk.h>
#include <zephyr/logging/log.h>
#include <stdio.h>
#include <stdlib.h>
#include <psa/crypto.h>
#include <psa/crypto_extra.h>

#include "key.h"

#ifdef CONFIG_BUILD_WITH_TFM
#include <tfm_ns_interface.h>
#endif

#define APP_SUCCESS	    (0)
#define APP_ERROR	    (-1)
#define APP_SUCCESS_MESSAGE "Example finished successfully!"
#define APP_ERROR_MESSAGE   "Example exited with error!"

#define PRINT_HEX(p_label, p_text, len)                                                            \
	({                                                                                         \
		LOG_INF("---- %s (len: %u): ----", p_label, len);                                  \
		LOG_HEXDUMP_INF(p_text, len, "Content:");                                          \
		LOG_INF("---- %s end  ----", p_label);                                             \
	})

LOG_MODULE_REGISTER(rsa, LOG_LEVEL_DBG);

//### ADDED TO SAMPLE - START
uint8_t encrypted[294];

/* ====================================================================== */
/*				Global variables/defines for the RSA example			  */

// #ifndef CONFIG_PSA_WANT_RSA_KEY_SIZE_4096
// #error "This sample needs a key size of 4096"
// #endif

#define NRF_CRYPTO_EXAMPLE_RSA_TEXT_SIZE       (100)
#define NRF_CRYPTO_EXAMPLE_RSA_PUBLIC_KEY_SIZE (PSA_KEY_EXPORT_RSA_PUBLIC_KEY_MAX_SIZE(2048))
#define NRF_CRYPTO_EXAMPLE_RSA_SIGNATURE_SIZE  (PSA_BITS_TO_BYTES(2048))

#define PSA_ALG_RSA_OAEP_EN 1
#define PSA_ALG_RSA_PKCS1V15_CRYPT_EN 0

uint8_t key[16]={0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x0A,0x0B,0x0C,0x0D,0x0E,0x0F};

/* Below text is used as plaintext for signing using RSA . */
static char m_plain_text[NRF_CRYPTO_EXAMPLE_RSA_TEXT_SIZE] = {
	"Example string to demonstrate basic usage of RSA."};

static char m_signature[NRF_CRYPTO_EXAMPLE_RSA_SIGNATURE_SIZE];
static char m_hash[32];

static psa_key_id_t keypair_handle;
static psa_key_id_t pub_key_handle;
/* ====================================================================== */

int crypto_init(void)
{
	psa_status_t status;

	/* Initialize PSA Crypto */
	status = psa_crypto_init();
	if (status != PSA_SUCCESS) {
		return APP_ERROR;
	}

	return APP_SUCCESS;
}

int import_rsa_pub_key(void)
{
    /* 기존 키가 있다면 삭제 */
    if (pub_key_handle != 0) {
        psa_destroy_key(pub_key_handle);
        pub_key_handle = 0;
    }

    /* Configure the key attributes */
    psa_key_attributes_t key_attributes = PSA_KEY_ATTRIBUTES_INIT;
    psa_status_t status;

    /* ✅ PSA_KEY_USAGE_ENCRYPT 강제 적용 */
    psa_key_usage_t usage_flags = PSA_KEY_USAGE_ENCRYPT | PSA_KEY_USAGE_VERIFY_MESSAGE | PSA_KEY_USAGE_EXPORT;
    psa_set_key_usage_flags(&key_attributes, usage_flags);
    psa_set_key_lifetime(&key_attributes, PSA_KEY_LIFETIME_VOLATILE);

#if PSA_ALG_RSA_PKCS1V15_CRYPT_EN
	psa_set_key_algorithm(&key_attributes, PSA_ALG_RSA_PKCS1V15_CRYPT);
#endif
#if PSA_ALG_RSA_OAEP_EN
	psa_set_key_algorithm(&key_attributes, PSA_ALG_RSA_OAEP(PSA_ALG_SHA_256));
#endif

    psa_set_key_type(&key_attributes, PSA_KEY_TYPE_RSA_PUBLIC_KEY);
    psa_set_key_bits(&key_attributes, 2048);

    LOG_INF("Public key length = %d", sizeof(public_key_der));

    status = psa_import_key(&key_attributes, public_key_der, sizeof(public_key_der), &pub_key_handle);
    if (status != PSA_SUCCESS) {
        LOG_INF("psa_import_key failed! (Error: %d)", status);
        return APP_ERROR;
    }

    return APP_SUCCESS;
}

bool rsaEncrypt(uint8_t *plainData, uint16_t plainDataLength, uint8_t *encDataOut, uint16_t *encDataLen)
{
	psa_status_t status;
	uint32_t outputLength = 0;
	psa_key_attributes_t key_attr;

	// RSA 키 속성 가져오기
	psa_get_key_attributes(pub_key_handle, &key_attr);
	size_t rsa_key_size = PSA_BITS_TO_BYTES(psa_get_key_bits(&key_attr));

	// 디버깅 로그 추가
	LOG_INF("RSA Key Size (bytes): %d", rsa_key_size);
	LOG_INF("Encryption Buffer Size: %d", *encDataLen);

	// 키 크기 확인
	if (psa_get_key_bits(&key_attr) < 2048) {
		LOG_INF("RSA key size too small! Expected at least 2048 bits.");
		return PSA_ERROR_INVALID_ARGUMENT;
	}

#if PSA_ALG_RSA_OAEP_EN
	/* 🔍 키가 OAEP를 지원하는지 확인 */
	psa_algorithm_t key_alg = psa_get_key_algorithm(&key_attr);
	LOG_INF("Key Algorithm After Import: 0x%x", key_alg);

	if (key_alg != PSA_ALG_RSA_OAEP(PSA_ALG_SHA_256)) {
		LOG_INF("Error: Imported key does not support RSA_OAEP(SHA-256)");
		return PSA_ERROR_NOT_SUPPORTED;
	}


	// RSA-OAEP 최대 입력 데이터 크기 계산 (SHA-256 사용)
	size_t max_plaintext_size = rsa_key_size - (2 * 32) - 2;
	LOG_INF("Max Plaintext Size: %d, Provided Plaintext Size: %d", max_plaintext_size, plainDataLength);

	// 입력 크기 초과 확인
	if (plainDataLength > max_plaintext_size) {
		LOG_INF("Input data too large for RSA-OAEP!");
		return PSA_ERROR_INVALID_ARGUMENT;
	}
#endif

	// 키 타입 확인
	if (psa_get_key_type(&key_attr) != PSA_KEY_TYPE_RSA_PUBLIC_KEY) {
		LOG_INF("Incorrect key type! Expected RSA public key.");
		return PSA_ERROR_INVALID_ARGUMENT;
	}

	// 키 사용 권한 확인
	psa_key_usage_t usage_flags = psa_get_key_usage_flags(&key_attr);
	if ((usage_flags & PSA_KEY_USAGE_ENCRYPT) == 0) {
		LOG_INF("Key does not have encryption usage flag!");
		return PSA_ERROR_NOT_PERMITTED;
	}

	LOG_INF("Key Type: 0x%x", psa_get_key_type(&key_attr));
	LOG_INF("Key Algorithm: 0x%x", psa_get_key_algorithm(&key_attr));
	LOG_INF("Key Usage Flags: 0x%x", psa_get_key_usage_flags(&key_attr));
	LOG_INF("Key Bits: %d", psa_get_key_bits(&key_attr));

#if PSA_ALG_RSA_OAEP_EN
	// Label 길이 문제 해결 → NULL 사용
	status = psa_asymmetric_encrypt(pub_key_handle,
					PSA_ALG_RSA_OAEP(PSA_ALG_SHA_256),
					plainData,
					plainDataLength,
					NULL,  // Label을 NULL로 설정
					0,     // Label 길이를 0으로 설정
					encDataOut,
					*encDataLen,
					&outputLength);
#endif



#if PSA_ALG_RSA_PKCS1V15_CRYPT_EN

	// Label 길이 문제 해결 → NULL 사용
	status = psa_asymmetric_encrypt(pub_key_handle,
		PSA_ALG_RSA_PKCS1V15_CRYPT,
		plainData,
		plainDataLength,
		NULL,  // Label을 NULL로 설정
		0,     // Label 길이를 0으로 설정
		encDataOut,
		*encDataLen,
		&outputLength);
#endif

	if (status != PSA_SUCCESS) {
		LOG_INF("Encryption failed! Status: %d", status);
	} else {
		LOG_INF("Encryption successful! Output Length: %d", outputLength);
	}

	PRINT_HEX("Encrypted Data", encDataOut, *encDataLen);

	return status;
}


int testEncryption() {
    uint8_t encDataOut[256];  // 충분한 크기의 버퍼 할당
    uint16_t encDataLen = sizeof(encDataOut);  // 버퍼 크기를 설정

    int result = rsaEncrypt(key, sizeof(key), encDataOut, &encDataLen);
    if (!result) {
        LOG_INF("Encryption failed: %d\n", result);
        return APP_ERROR;
    }
	
    return APP_SUCCESS;
}

int crypto_finish(void)
{
	psa_status_t status;

	/* Destroy the key handle */
	status = psa_destroy_key(keypair_handle);
	if (status != PSA_SUCCESS) {
		LOG_INF("psa_destroy_key failed! (Error: %d)", status);
		return APP_ERROR;
	}

	status = psa_destroy_key(pub_key_handle);
	if (status != PSA_SUCCESS) {
		LOG_INF("psa_destroy_key failed! (Error: %d)", status);
		return APP_ERROR;
	}

	return APP_SUCCESS;
}

int main(void)
{
	int status;

	LOG_INF("Starting the RSA example...");

	status = crypto_init();
	if (status != APP_SUCCESS) {
		// LOG_INF(APP_ERROR_MESSAGE);
		return APP_ERROR;
	}
	status = import_rsa_pub_key();
	if (status != APP_SUCCESS) {
		// LOG_INF(APP_ERROR_MESSAGE);
		return APP_ERROR;
	}

	status = testEncryption();
	if (status != APP_SUCCESS) {
		// LOG_INF(APP_ERROR_MESSAGE);
		return APP_ERROR;
	}

	// k_msleep(1000); 

	// while (1)
	// {
	// 	for(int i = 0; i < 1; i++)
	// 	{
	// 		status = testEncryption();
	// 		if (status != APP_SUCCESS) {
	// 			LOG_INF(APP_ERROR_MESSAGE);
	// 			return APP_ERROR;
	// 		}
	// 	}

	// 	k_msleep(1000); 
	// }

	// status = crypto_finish();
	// if (status != APP_SUCCESS) {
	// 	LOG_INF(APP_ERROR_MESSAGE);
	// 	return APP_ERROR;
	// }

	return APP_SUCCESS;
}
